# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys

from contextlib import suppress
from typing import Any, NoReturn, TextIO

from hatch_fancy_pypi_readme.exceptions import ConfigurationError

from ._builder import build_text
from ._config import load_and_validate_config


def cli_run(
    pyproject: dict[str, Any], hatch_toml: dict[str, Any], out: TextIO
) -> None:
    """
    Best-effort verify config and print resulting PyPI readme.
    """
    is_dynamic = False
    with suppress(KeyError):
        is_dynamic = "readme" in pyproject["project"]["dynamic"]

    if not is_dynamic:
        _fail("You must add 'readme' to 'project.dynamic'.")

    try:
        if (
            pyproject["tool"]["hatch"]["metadata"]["hooks"][
                "fancy-pypi-readme"
            ]
            and hatch_toml["metadata"]["hooks"]["fancy-pypi-readme"]
        ):
            _fail(
                "Both pyproject.toml and hatch.toml contain "
                "hatch-fancy-pypi-readme configuration."
            )
    except KeyError:
        pass

    try:
        cfg = hatch_toml["metadata"]["hooks"]["fancy-pypi-readme"]
    except KeyError:
        try:
            cfg = pyproject["tool"]["hatch"]["metadata"]["hooks"][
                "fancy-pypi-readme"
            ]
        except KeyError:
            _fail(
                "Missing configuration "
                "(`[tool.hatch.metadata.hooks.fancy-pypi-readme]` in"
                " pyproject.toml or `[metadata.hooks.fancy-pypi-readme]`"
                " in hatch.toml)",
            )

    try:
        config = load_and_validate_config(cfg)
    except ConfigurationError as e:
        _fail(
            "Configuration has errors:\n\n"
            + "\n".join(f"- {msg}" for msg in e.errors),
        )

    print(build_text(config.fragments, config.substitutions, "42.0"), file=out)


def _fail(msg: str) -> NoReturn:
    print(msg, file=sys.stderr)
    sys.exit(1)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any

from hatchling.metadata.plugin.interface import MetadataHookInterface
from hatchling.plugin import hookimpl

from ._builder import build_text
from ._config import load_and_validate_config


class FancyReadmeMetadataHook(MetadataHookInterface):
    PLUGIN_NAME = "fancy-pypi-readme"

    def update(self, metadata: dict[str, Any]) -> None:
        """
        Update the project table's metadata.
        """
        config = load_and_validate_config(self.config)

        metadata["readme"] = {
            "content-type": config.content_type,
            "text": build_text(
                config.fragments,
                config.substitutions,
                version=metadata.get("version", ""),
            ),
        }


@hookimpl
def hatch_register_metadata_hook() -> type[MetadataHookInterface]:
    return FancyReadmeMetadataHook



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re
import sys

from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar, Iterable


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from .exceptions import ConfigurationError


class Fragment(Protocol):
    key: ClassVar[str]

    @classmethod
    def from_config(cls, cfg: dict[str, str]) -> Fragment:
        ...

    def render(self) -> str:
        ...


@dataclass
class TextFragment:
    """
    A static text fragment.
    """

    key: ClassVar[str] = "text"

    _text: str

    @classmethod
    def from_config(cls, cfg: dict[str, str]) -> Fragment:
        text = cfg[cls.key]
        if not text:
            raise ConfigurationError(["Text fragments must not be empty."])

        return cls(text)

    def render(self) -> str:
        return self._text


@dataclass
class FileFragment:
    """
    A static text fragment.
    """

    key: ClassVar[str] = "path"

    _contents: str

    @classmethod
    def from_config(cls, cfg: dict[str, str]) -> Fragment:
        path = Path(cfg.pop(cls.key))
        start_after = cfg.pop("start-after", None)
        start_at = cfg.pop("start-at", None)
        end_before = cfg.pop("end-before", None)
        pattern = cfg.pop("pattern", None)

        errs: list[str] = []

        try:
            contents = path.read_text(encoding="utf-8")
        except FileNotFoundError:
            raise ConfigurationError(
                [f"Fragment file '{path}' not found."]
            ) from None

        if start_after and start_at:
            raise ConfigurationError(
                [
                    "file fragment: 'start-after' and 'start-at' are "
                    "mutually exclusive."
                ]
            )

        if start_after is not None:
            try:
                _, contents = contents.split(start_after, 1)
            except ValueError:
                errs.append(
                    f"file fragment: 'start-after' {start_after!r} not found."
                )
        elif start_at is not None:
            p = contents.find(start_at)
            if p == -1:
                errs.append(
                    f"file fragment: 'start-at' {start_at!r} not found."
                )
            contents = contents[p:]

        if end_before is not None:
            try:
                contents, _ = contents.split(end_before, 1)
            except ValueError:
                errs.append(
                    f"file fragment: 'end-before' {end_before!r} not found."
                )

        if pattern:
            m = re.search(pattern, contents, re.DOTALL)
            if not m:
                errs.append(f"file fragment: pattern {pattern!r} not found.")
            else:
                try:
                    contents = m.group(1)
                except IndexError:
                    errs.append(
                        "file fragment: pattern matches, but no group "
                        "defined."
                    )

        if errs:
            raise ConfigurationError(errs)

        return cls(contents)

    def render(self) -> str:
        return self._contents


VALID_FRAGMENTS: Iterable[type[Fragment]] = (TextFragment, FileFragment)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ._fragments import Fragment
    from ._substitutions import Substituter


def build_text(
    fragments: list[Fragment],
    substitutions: list[Substituter],
    version: str = "",
) -> str:
    """
    Try avoiding breaking the API unnecessarily; it's used directly by
    scikit-build-core.
    """
    text = "".join(f.render() for f in fragments)

    for sub in substitutions:
        text = sub.substitute(text)

    return text.replace("$HFPR_VERSION", version)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re

from dataclasses import dataclass
from typing import cast

from hatch_fancy_pypi_readme.exceptions import ConfigurationError


@dataclass
class Substituter:
    pattern: re.Pattern[str]
    replacement: str

    @classmethod
    def from_config(cls, cfg: dict[str, str]) -> Substituter:
        errs = []
        flags = 0

        ignore_case = cfg.get("ignore-case", False)
        if not isinstance(ignore_case, bool):
            errs.append(
                f"Value {ignore_case!r} for 'ignore-case' is not a bool."
            )

        if ignore_case:
            flags += re.IGNORECASE

        try:
            pattern = re.compile(cfg["pattern"], flags=flags)
        except KeyError:
            errs.append(f"Substitution {cfg} is missing a 'pattern' key.")
        except re.error as e:
            errs.append(
                f"{cfg['pattern']!r} is not a valid regular expression: {e}"
            )

        replacement = cfg.get("replacement")
        if replacement is None:
            errs.append(f"Substitution {cfg} is missing a 'replacement' key.")
        elif not isinstance(replacement, str):
            errs.append(f"Replacement value {replacement!r} is not a string.")

        if errs:
            raise ConfigurationError(errs)

        return cls(pattern, cast(str, replacement))

    def substitute(self, text: str) -> str:
        return self.pattern.sub(self.replacement, text)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import argparse
import sys

from contextlib import closing
from pathlib import Path
from typing import TextIO

from ._cli import cli_run


if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Render a README from a pyproject.toml & hatch.toml."
        " If a hatch.toml is passed / detected, it's preferred."
    )
    parser.add_argument(
        "pyproject_path",
        nargs="?",
        metavar="PATH-TO-PYPROJECT.TOML",
        default="pyproject.toml",
        help="Path to the pyproject.toml to use for rendering. "
        "Default: pyproject.toml in current directory.",
    )
    parser.add_argument(
        "--hatch-toml",
        nargs="?",
        metavar="PATH-TO-HATCH.TOML",
        default=None,
        help="Path to an additional hatch.toml to use for rendering. "
        "Default: Auto-detect in the current directory.",
    )
    parser.add_argument(
        "-o",
        help="Target file for output. Default: standard out.",
        metavar="TARGET-FILE-PATH",
    )
    args = parser.parse_args()

    pyproject = tomllib.loads(Path(args.pyproject_path).read_text())
    hatch_toml = _maybe_load_hatch_toml(args.hatch_toml)

    out: TextIO
    out = Path(args.o).open("w") if args.o else sys.stdout  # noqa: SIM115

    with closing(out):
        cli_run(pyproject, hatch_toml, out)


def _maybe_load_hatch_toml(hatch_toml_arg: str | None) -> dict[str, object]:
    """
    If hatch.toml is passed or detected, load it.
    """
    if hatch_toml_arg:
        return tomllib.loads(Path(hatch_toml_arg).read_text())

    if Path("hatch.toml").exists():
        return tomllib.loads(Path("hatch.toml").read_text())

    return {}


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class ConfigurationError(Exception):
    """
    Configuration is invalid.
    """

    errors: list[str]



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, cast

from ._fragments import VALID_FRAGMENTS, Fragment
from ._substitutions import Substituter
from .exceptions import ConfigurationError


@dataclass
class Config:
    content_type: str
    fragments: list[Fragment]
    substitutions: list[Substituter]


_BASE = "tool.hatch.metadata.hooks.fancy-pypi-readme."


def load_and_validate_config(config: dict[str, Any]) -> Config:
    errs = []

    ct = config.get("content-type")
    if ct is None:
        errs.append(f"{_BASE}content-type is missing.")
    elif ct not in ("text/markdown", "text/x-rst"):
        errs.append(
            f"{_BASE}content-type: '{ct}' is not one of "
            "['text/markdown', 'text/x-rst']"
        )

    try:
        fragments = _load_fragments(config.get("fragments"))
    except ConfigurationError as e:
        errs.extend(e.errors)

    try:
        subs_cfg = config.get("substitutions", [])
        if not isinstance(subs_cfg, list):
            raise ConfigurationError(
                [f"{_BASE}substitutions must be an array."]
            )

        substitutions = [
            Substituter.from_config(sub_cfg) for sub_cfg in subs_cfg
        ]
    except ConfigurationError as e:
        errs.extend(e.errors)

    if errs:
        raise ConfigurationError(errs)

    return Config(
        content_type=cast(str, ct),
        fragments=fragments,
        substitutions=substitutions,
    )


def _load_fragments(config: list[dict[str, str]] | None) -> list[Fragment]:
    """
    Load fragments from *config*.
    """
    if config is None:
        raise ConfigurationError([f"{_BASE}fragments is missing."])
    if not config:
        raise ConfigurationError([f"{_BASE}fragments must not be empty."])

    frags = []
    errs = []

    for frag_cfg in config:
        for frag in VALID_FRAGMENTS:
            if frag.key not in frag_cfg:
                continue

            try:
                frags.append(frag.from_config(frag_cfg))
            except ConfigurationError as e:
                errs.extend(e.errors)

            # We have either detected and added or detected and errored, but in
            # any case we're done with this fragment.
            break
        else:
            errs.append(f"Unknown fragment type {frag_cfg!r}.")

    if errs:
        raise ConfigurationError(errs)

    return frags
